/*
 * Souffle - A Datalog Compiler
 * Copyright (c) 2021, The Souffle Developers. All rights reserved
 * Licensed under the Universal Permissive License v 1.0 as shown at:
 * - https://opensource.org/licenses/UPL
 * - <souffle root>/licenses/SOUFFLE-UPL.txt
 */

/************************************************************************
 *
 * @file AbstractParallel.h
 *
 ***********************************************************************/

#pragma once

namespace souffle::ram {

/**
 * @class AbstractParallel
 * @brief Abstract class for parallel operation
 */
class AbstractParallel {};

}  // namespace souffle::ram
