/*
 * Souffle - A Datalog Compiler
 * Copyright (c) 2021, The Souffle Developers. All rights reserved
 * Licensed under the Universal Permissive License v 1.0 as shown at:
 * - https://opensource.org/licenses/UPL
 * - <souffle root>/licenses/SOUFFLE-UPL.txt
 */

/************************************************************************
 *
 * @file lib_foo.cpp
 *
 * See `lib_common.h` for details.
 *
 ***********************************************************************/

#define DEFINE_FOO
#include "lib_common.h"
